# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join("..", "src", "semantic_world")))

# -- Project information -----------------------------------------------------

project = 'semantic_world'
copyright = '2025, AICOR Institute for Artificial Intelligence'
author = 'AICOR Institute for Artificial Intelligence'

# The short X.Y version
version = ''
# The full version, including alpha/beta/rc tags
release = ''